from cel.connectors.whatsapp.phone_utils import filter_phone_number
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.conversation_peer import ConversationPeer


class WhatsappLead(ConversationLead):

    def __init__(self, phone: str, connector_name: str = None, **kwargs):
        super().__init__(connector_name=connector_name or 'whatsapp', **kwargs)
        self.phone: str = filter_phone_number(str(phone))

    def get_session_id(self):
        return f"{self.connector_name}:{self.phone}"  

    def to_dict(self):
        data = super().to_dict()
        data['phone'] = self.phone
        return data

    @classmethod
    def from_dict(cls, lead_dict):
        return WhatsappLead(
            phone=lead_dict.get("phone"),
            metadata=lead_dict.get("metadata")
        )

    def __str__(self):
        return f"WhatsappLead: {self.phone}"
    
    @classmethod
    def from_whatsapp_message(cls, data: dict, **kwargs):
        assert isinstance(data, dict), "data must be a dictionary"
        
        phone = data.get("entry")[0].get("changes")[0].get("value").get("contacts")[0].get("wa_id")
        
        metadata = {
            'phone_number_id': data.get("entry")[0].get("changes")[0].get("value").get("metadata").get("phone_number_id"),
            'display_phone_number': data.get("entry")[0].get("changes")[0].get("value").get("metadata").get("display_phone_number"),
            'message_id': data.get("entry")[0].get("id"),
            'date': data.get("entry")[0].get("changes")[0].get("value").get("timestamp"),
            'raw': data,
            'wamid': data.get("entry")[0].get("changes")[0].get("value").get("messages")[0].get("id")
        }
        conversation_peer = ConversationPeer(
            name=data.get("entry")[0].get("changes")[0].get("value").get("contacts")[0].get("profile").get("name"),
            id=data.get("entry")[0].get("changes")[0].get("value").get("contacts")[0].get("wa_id"),
            phone=phone,
            avatarUrl=None,
            email=None
        )
        return WhatsappLead(phone=phone, metadata=metadata, conversation_from=conversation_peer, **kwargs)


from cel.gateway.model.conversation_lead import ConversationLead
from langchain.load.dump import dumpd
from langchain.load.load import load
from langchain_core.messages import BaseMessage
from cel.stores.history.base_history_provider import BaseHistoryProvider


class MacawHistoryAdapter:
    def __init__(self, store: BaseHistoryProvider):
        self.store = store

    async def append_to_history(self, lead: ConversationLead, entry: BaseMessage, metadata=None, ttl=None):
        assert isinstance(lead, ConversationLead), f"Expected ConversationLead, got {type (lead)}"
        aux = dumpd(entry)
        await self.store.append_to_history(lead.get_session_id(), aux, metadata, ttl)

    async def get_history(self, lead: ConversationLead) -> list[BaseMessage]:
        assert isinstance(lead, ConversationLead), f"Expected ConversationLead, got {type (lead)}"
        history = await self.store.get_history(lead.get_session_id())
        return [load(h) for h in history]


    async def clear_history(self, lead: ConversationLead, keep_last_messages=None):
        assert isinstance(lead, ConversationLead), f"Expected ConversationLead, got {type (lead)}"
        await self.store.clear_history(lead.get_session_id(), keep_last_messages)

    async def get_last_messages(self, lead: ConversationLead, count) -> list[BaseMessage]:
        assert isinstance(lead, ConversationLead), f"Expected ConversationLead, got {type (lead)}"
        msgs = await self.store.get_last_messages(lead.get_session_id(), count)
        return [load(m) for m in msgs]

    async def close_conversation(self, lead: ConversationLead):
        raise NotImplementedError


from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path



#  A Slice has a text, metadata: dict, and a id: str, source: str
@dataclass
class Slice(ABC):
    """
    A class used to represent a Slice.

    Attributes
    ----------
    id : str
        unique identifier for the slice
    text : str
        text content of the slice
    metadata : dict
        additional information about the slice
    source : str
        source from which the slice was derived
    """
    id: str
    text: str
    metadata: dict
    source: str    
    
    

class Slicer(ABC):
        
    @abstractmethod
    def slice(self) -> list[Slice]:
        pass
    
    



import json
from redis import Redis
from loguru import logger as log
from cel.stores.state.base_state_provider import BaseChatStateProvider

class RedisChatStateProvider(BaseChatStateProvider):

    def __init__(self, redis: str | Redis, key_prefix: str = "s"):
        super().__init__()
        log.debug("Create: RedisChatStateProvider")
        self.client = redis if isinstance(redis, Redis) else Redis.from_url(redis)
        self.prefix = key_prefix
        
    def get_key(self, sessionId):
        return f"{self.prefix}:{sessionId}"
        
    def set_key_value(self, sessionId: str, key: str, value, ttl_in_seconds=None):
        hash_key = self.get_key(sessionId)
        self.client.hset(hash_key, key, json.dumps(value))
        if ttl_in_seconds:
            self.client.expire(hash_key, ttl_in_seconds)

    def get_key_value(self, sessionId: str, key: str):
        hash_key = self.get_key(sessionId)
        value = self.client.hget(hash_key, key)
        if not value:
            return None
        return json.loads(value)

    def clear_store(self, sessionId: str):
        hash_key = self.get_key(sessionId)
        self.client.delete(hash_key)

    def clear_all_stores(self):
        hash_key = self.get_key("*")
        keys = self.client.keys(hash_key)
        if not keys:
            return
        for key in keys:
            self.client.delete(key)

    def get_store(self, sessionId: str):
        hash_key = self.get_key(sessionId)
        store = self.client.hgetall(hash_key)
        if not store:
            return None
        s = {k.decode('utf-8'): json.loads(store[k]) for k in store}
        return s


    def set_store(self, sessionId: str, store, ttl=None):
        hash_key = self.get_key(sessionId)
        for key in store:
            self.client.hset(hash_key, key, json.dumps(store[key]))
        if ttl:
            self.client.expire(hash_key, ttl)





from abc import ABC
import os
import time
from typing import cast
from diskcache import Cache
from cel.cache import get_cache
from .utils import Embedding, Embeddings, Text2VectorProvider

try:
    from openai import OpenAI
except ImportError:
    raise ValueError(
        "The openai python package is not installed. Please install it with `pip install openai`"
    )
    
cache = get_cache()
CACHE_TAG = 'oaie'
CACHE_EXPIRE = None
DEFAULT_MAX_RETRIES = 5

class CachedOpenAIEmbedding(Text2VectorProvider):
    """A wrapper around the OpenAI API that caches the results of text2vec calls.
    Uses diskcache to cache the results of text2vec calls.
    Usefull for reducing the number of API calls to OpenAI Embedding API.
    For example, if you are using ChromaDB every time you boot up your application, 
    Chroma needs to be re-embedded. This can be time consuming and costly.
    This class will cache the results of the text2vec calls, so that the next time you boot up your application,
    the embeddings will be retrieved from the cache.

    
    Parameters:
    api_key: str
        The OpenAI API key. If not provided, it will look for the OPENAI_API_KEY environment variable.
    model: str
        The OpenAI model to use. Default is "text-embedding-3-small"
    max_retries: int
        The maximum number of retries to make when the API call fails. Default is 5.
    kwargs: dict
    """
    
    def __init__(self, api_key: str = None, model: str = "text-embedding-3-small", max_retries: int = DEFAULT_MAX_RETRIES):
        self.api_key = api_key or os.getenv("OPENAI_API_KEY")
        self.model = model

        if self.api_key is not None:
            OpenAI.api_key = api_key
        # If the api key is still not set, raise an error
        elif OpenAI.api_key is None:
            raise ValueError(
                "Please provide an OpenAI API key. You can get one at https://platform.openai.com/account/api-keys"
            )
            
        
        self.max_retries = max_retries

    def text2vec(self, text: str) -> Embedding:
        return openai_cached_text2vec(text, self.model, self.max_retries)
    
    def texts2vec(self, texts: list[str]) -> list[Embedding]:
        return [self.text2vec(text) for text in texts]
    
    
    
    
@cache.memoize(typed=True, expire=CACHE_EXPIRE, tag=CACHE_TAG)
def openai_cached_text2vec(text: str, model: str, max_retries: int = 3) -> list[float]:
    
    client = OpenAI(max_retries=max_retries)

    # replace newlines, which can negatively affect performance.
    text = text.replace("\n", " ")

    response = client.embeddings.create(input=[text], model=model)
    embeddings = response.data[0].embedding
    return embeddings
    




import asyncio
import inspect
import json
import re
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message


class PromptTemplate:
    def __init__(self, prompt: str):
        self.prompt = prompt

    async def compile(self, state: dict, lead: ConversationLead):
        assert isinstance(state, dict), "PromptTemplate: state must be a dictionary"
        
        async def compile_value(key):
            var_name = key
            if var_name in state:
                value = state[var_name]
                if callable(value):
                    return { key: await self.call_function(value, lead=lead) }
                else:
                    return { key: str(value) }
            else:
                return { key: '' }

        pattern = re.compile(r'\{(\w+)\}')
        matches = re.findall(pattern, self.prompt)
        # Asumiendo que 'matches' es la lista de coincidencias
        tasks = [asyncio.ensure_future(compile_value(match)) for match in matches]
        await asyncio.gather(*tasks)

        prompt = self.prompt
        for task in tasks:
            # replace the key with the value in the prompt
            key, value = task.result().popitem()
            prompt = prompt.replace(f"{{{key}}}", value)
        return prompt
            

    async def call_function(self, 
                            func: callable, 
                            message: Message = None,
                            lead: ConversationLead = None) -> str:
        
        args_dict = {
            'lead': lead,
            'session_id': lead.get_session_id(),
            'message': message
        }

        args = inspect.getfullargspec(func).args
        # Build kwargs using args and args_dict
        kwargs = {}
        for arg in args:
            if arg in args_dict:
                kwargs[arg] = args_dict[arg]

        # Check if the function is a coroutine (async function)
        response = None
        if inspect.iscoroutinefunction(func):
            response = await func(**kwargs)
        else:
            response = func(**kwargs)

        return json.dumps(response)


# https://medium.com/@gal.peretz/openrouter-langchain-leverage-opensource-models-without-the-ops-hassle-9ffbf0016da7

import os
from typing import Optional
# from langchain_community.chat_models import ChatOpenAI
from langchain_openai import ChatOpenAI
from langchain_core.prompts import ChatPromptTemplate


OPENROUTER_API_BASE = "https://openrouter.ai/api/v1"


# TODO: OpenRouter Models has problems with Langchain tooling
# This LLM model wrapper throws an error when used with Langchain tooling
# specifically with bind_tools. The error is due to the fact that the
# OpenRouter models do not support tooling??? I am not sure about this.
# --------------------------------------------------------------------
class ChatOpenRouter(ChatOpenAI):
    openai_api_base: str
    openai_api_key: str
    model_name: str

    def __init__(self,
                 model: str,
                 openai_api_key: Optional[str] = None,
                 openai_api_base: str = OPENROUTER_API_BASE,
                 **kwargs):
        openai_api_key = openai_api_key or os.getenv('OPENROUTER_API_KEY')
        super().__init__(openai_api_base=openai_api_base,
                         openai_api_key=openai_api_key,
                         model_name=model, **kwargs)
        
      
def ChatOpenAIOpenRouter(**kwargs):
    return ChatOpenAI(
        **kwargs, 
        openai_api_base=OPENROUTER_API_BASE,
        openai_api_key = os.getenv('OPENROUTER_API_KEY')
    )
# --------------------------------------------------------------------
        




from cel.gateway.model.attachment import FileAttachment
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.outgoing.outgoing_message import OutgoingMessage, OutgoingMessageType


class OutgoingSelectMessage(OutgoingMessage):
    """This class represents a specific outgoing message object with select options"""
    
    def __init__(self, 
                 lead: ConversationLead, 
                 metadata: dict = None, 
                 attachments: list[FileAttachment] = None,
                 is_partial: bool = True,
                 content: str = None,
                 options: list[str] = None
                ):
        super().__init__(OutgoingMessageType.SELECT, lead, metadata, attachments, is_partial)
        self.content = content
        self.options = options

        assert isinstance(self.content, str), "body must be a string"
        assert isinstance(self.options, list), "options must be a list"
        for option in self.options:
            assert isinstance(option, str), "each option must be a string"
            
            
    @staticmethod
    def from_dict(data: dict) -> 'OutgoingSelectMessage':
        """Creates an OutgoingSelectMessage instance from a dictionary"""
        assert isinstance(data, dict),\
            "data must be a dictionary"
        assert "content" in data,\
            "data must have a 'content' key"
        assert "options" in data,\
            "data must have a 'options' key"
        assert isinstance(data["options"], list),\
            "options must be a list"
        assert "lead" in data,\
            "data must have a 'lead' key"
        assert isinstance(data["lead"], ConversationLead),\
            "lead must be an instance of ConversationLead"
        
        return OutgoingSelectMessage(
            content=data["content"],
            options=data["options"],
            lead=data["lead"],
            metadata=data.get("metadata"),
            attachments=[FileAttachment.from_dict(attachment) for attachment in data.get("attachments", [])],
            is_partial=data.get("is_partial", True)
        )
        
        
    @staticmethod
    def description() -> str:
        return """When the message asks the user to make a choice, use the following structure:
{
    "type": "select",
    "options": [],
    "content": "A boody text",
}

The options must be a list of strings. Options should be short and clear in title case. 
The prompt is keeped in the content field, and the options are the possible choices.
"""


from abc import ABC
import os
import time
from typing import cast
from diskcache import Cache

from cel.cache import get_cache
from .utils import Embedding, Embeddings, Text2VectorProvider

try:
    import ollama
except ImportError:
    raise ValueError(
        "The Ollama library is required for this example. Please install it by running: pip install ollama"
    )
    
cache = get_cache()
CACHE_TAG = 'ollama'
CACHE_EXPIRE = None
DEFAULT_MAX_RETRIES = 5

class CachedOllamaEmbedding(Text2VectorProvider):
    """A wrapper around the Ollama library that caches the results of the embeddings calls.
    Uses diskcache to cache the results of text2vec calls.
    Usefull for reducing the number of API calls
    For example, if you are using ChromaDB every time you boot up your application, 
    Chroma needs to be re-embedded. This can be time consuming and costly.
    This class will cache the results of the text2vec calls, so that the next time you boot up your application,
    the embeddings will be retrieved from the cache.

    
    Parameters:
    model: str
        Ollama embedding models, check the Ollama documentation for the available models.
        https://ollama.com/library
        default: "mxbai-embed-large"
    """
    
    def __init__(self, model: str = "mxbai-embed-large"):
        self.model = model


    def text2vec(self, text: str) -> Embedding:
        return ollama_cached_text2vec(text, self.model)
    
    def texts2vec(self, texts: list[str]) -> list[Embedding]:
        return [self.text2vec(text) for text in texts]
    
    
    
    
@cache.memoize(typed=True, expire=CACHE_EXPIRE, tag=CACHE_TAG)
def ollama_cached_text2vec(text: str, model: str) -> list[float]:
    response = ollama.embeddings(model=model, prompt=text)
    embedding = response["embedding"]
    return embedding
    



import logging
import requests
from typing import Dict, Any


def create_button(button: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Method to create a button object to be used in the send_message method.

    This is method is designed to only be used internally by the send_button method.

    Args:
            button[dict]: A dictionary containing the button data
    """
    data = {"type": "list", "action": button.get("action")}
    if button.get("header"):
        data["header"] = {"type": "text", "text": button.get("header")}
    if button.get("body"):
        data["body"] = {"text": button.get("body")}
    if button.get("footer"):
        data["footer"] = {"text": button.get("footer")}
    return data


def send_button(self, button: Dict[Any, Any], recipient_id: str) -> Dict[Any, Any]:
    """
    Sends an interactive buttons message to a WhatsApp user

    Args:
        button[dict]: A dictionary containing the button data(rows-title may not exceed 20 characters)
        recipient_id[str]: Phone number of the user with country code wihout +

    check https://github.com/Neurotech-HQ/whatsapp#sending-interactive-reply-buttons for an example.
    """
    data = {
        "messaging_product": "whatsapp",
        "to": recipient_id,
        "type": "interactive",
        "interactive": create_button(button),
    }
    logging.info(f"Sending buttons to {recipient_id}")
    r = requests.post(self.url, headers=self.headers, json=data)
    if r.status_code == 200:
        logging.info(f"Buttons sent to {recipient_id}")
        return r.json()
    logging.info(f"Buttons not sent to {recipient_id}")
    logging.info(f"Status code: {r.status_code}")
    logging.info(f"Response: {r.json()}")
    return r.json()


def send_reply_button(
    self, button: Dict[Any, Any], recipient_id: str
) -> Dict[Any, Any]:
    """
    Sends an interactive reply buttons[menu] message to a WhatsApp user

    Args:
        button[dict]: A dictionary containing the button data
        recipient_id[str]: Phone number of the user with country code wihout +

    Note:
        The maximum number of buttons is 3, more than 3 buttons will rise an error.
    """
    if len(button["action"]["buttons"]) > 3:
        raise ValueError("The maximum number of buttons is 3.")
    
    data = {
        "messaging_product": "whatsapp",
        "recipient_type": "individual",
        "to": recipient_id,
        "type": "interactive",
        "interactive": button,
    }
    r = requests.post(self.url, headers=self.headers, json=data)
    if r.status_code == 200:
        logging.info(f"Reply buttons sent to {recipient_id}")
        return r.json()
    logging.info(f"Reply buttons not sent to {recipient_id}")
    logging.info(f"Status code: {r.status_code}")
    logging.info(f"Response: {r.json()}")
    return r.json()




def send_cta_url(
    self, cta_url: Dict[Any, Any], recipient_id: str
) -> Dict[Any, Any]:
    """
    Sends a call to action url message to a WhatsApp user

    Args:
        cta_url[dict]: A dictionary containing the cta url data
        recipient_id[str]: Phone number of the user with country code wihout +

    check
    
    """
    data = {
        "messaging_product": "whatsapp",
        "recipient_type": "individual",
        "to": recipient_id,
        "type": "interactive",
        "interactive": cta_url,
    }
    r = requests.post(self.url, headers=self.headers, json=data)
    if r.status_code == 200:
        logging.info(f"CTA URL sent to {recipient_id}")
        return r.json()
    logging.info(f"CTA URL not sent to {recipient_id}")
    logging.info(f"Status code: {r.status_code}")
    logging.info(f"Response: {r.json()}")
    return r.json()


from loguru import logger as log
import requests
import os
import mimetypes
from requests_toolbelt.multipart.encoder import MultipartEncoder
from typing import Union, Dict, Any
from cel.connectors.whatsapp.constants import BASE_URL
from cel.connectors.whatsapp.utils import build_headers


def upload_media(media: str, token: str, phone_number_id: str) -> Union[Dict[Any, Any], None]:
    """
    Uploads a media to the cloud api and returns the id of the media

    Args:
        media[str]: Path of the media to be uploaded
        token[str]: The Meta Access Token
        phone_number_id[str]: The Meta Phone Number Id

    ref: https://developers.facebook.com/docs/whatsapp/cloud-api/reference/media#
    """
    assert token is not None, "Token not provided"
    assert phone_number_id is not None, "Phone number id not provided"
    assert media is not None, "Media not provided"

    form_data = {
        "file": (
            media,
            open(os.path.realpath(media), "rb"),
            mimetypes.guess_type(media)[0],
        ),
        "messaging_product": "whatsapp",
        "type": mimetypes.guess_type(media)[0],
    }
    form_data = MultipartEncoder(fields=form_data)
    headers = build_headers(token)
    headers["Content-Type"] = form_data.content_type
    log.info(f"Content-Type: {form_data.content_type}")
    log.info(f"Uploading media {media}")
    r = requests.post(
        f"{BASE_URL}/{phone_number_id}/media",
        headers=headers,
        data=form_data,
    )
    if r.status_code == 200:
        log.info(f"Media {media} uploaded")
        return r.json()
    log.info(f"Error uploading media {media}")
    log.info(f"Status code: {r.status_code}")
    log.info(f"Response: {r.json()}")
    return None


def delete_media(media_id: str, token: str) -> Union[Dict[Any, Any], None]:
    """
    Deletes a media from the cloud api

    Args:
        media_id[str]: Id of the media to be deleted
        token[str]: The Meta Access Token
    """
    assert token is not None, "Token not provided"
    assert media_id is not None, "Media id not provided"
    
    log.info(f"Deleting media {media_id}")
    r = requests.delete(
        f"{BASE_URL}/{media_id}", headers=build_headers(token))
    if r.status_code == 200:
        log.info(f"Media {media_id} deleted")
        return r.json()
    log.info(f"Error deleting media {media_id}")
    log.info(f"Status code: {r.status_code}")
    log.info(f"Response: {r.json()}")
    return None


def query_media_url(media_id: str, token: str) -> Union[str, None]:
    """
    Query media url from media id obtained either by manually uploading media or received media

    Args:
        media_id[str]: Media id of the media
        token[str]: The Meta Access Token

    Returns:
        str: Media url
    """
    assert media_id is not None, "Media id not provided"
    assert BASE_URL is not None, "Base url not provided"
    assert token is not None, "Token not provided"

    log.info(f"Querying media url for {media_id}")
    r = requests.get(f"{BASE_URL}/{media_id}", headers=build_headers(token))
    if r.status_code == 200:
        log.info(f"Media url queried for {media_id}")
        url = r.json()["url"]
        return url
    log.info(f"Media url not queried for {media_id}")
    log.info(f"Status code: {r.status_code}")
    log.info(f"Response: {r.json()}")
    return None


def download_media(media_url: str, 
                   mime_type: str, 
                   token: str, 
                   file_path: str = "temp"
                  ) -> Union[str, None]:
    """
    Download media from media url obtained either by manually uploading media or received media

    Args:
        media_url[str]: Media url of the media
        mime_type[str]: Mime type of the media
        file_path[str]: Path of the file to be downloaded to. Default is "temp"
                        Do not include the file extension. It will be added automatically.
        token[str]: The Meta Access Token
        
    Returns:
        str: Media url
    """
    assert media_url is not None, "Media url not provided"
    assert mime_type is not None, "Mime type not provided"
    assert token is not None, "Token not provided"
    
    r = requests.get(media_url, headers=build_headers(token))
    content = r.content
    extension = mime_type.split("/")[1]
    save_file_here = None
    # create a temporary file
    try:

        save_file_here = (
            f"{file_path}.{extension}" if file_path else f"temp.{extension}"
        )
        with open(save_file_here, "wb") as f:
            f.write(content)
        log.info(f"Media downloaded to {save_file_here}")
        return f.name
    except Exception as e:
        log.info(e)
        log.error(f"Error downloading media to {save_file_here}")
        return None



#   {
#     "type": "document",
#     "document": {
#       "id": "your-media-id",
#       # filename is an optional parameter
#       "filename": "your-document-filename"
#     }
#   }
from abc import ABC
import json
from .component import Component
from .utils import del_none

class Document(Component):
    def __init__(self, id: str, filename: str | None = None) -> None:
        self.id = id
        self.filename = filename

    def __str__(self):
        document = {
            "type": "document",
            "document": {
                "id": self.id,
                "filename": self.filename
            }
        }
        del_none(document)
        return json.dumps(document)

    def __repr__(self):
        return f"DocumentComponent(id={self.id}, filename={self.filename})"

    def __eq__(self, other):
        return self.id == other.id and self.filename == other.filename
    
    
if __name__ == "__main__":
    document = Document("your-media-id", "your-document-filename")
    print(document)
    # Expected output: '{"type": "document", "document": {"id": "your-media-id", "filename": "your-document-filename"}}'
    document = Document("your-media-id")
    print(document)    
    # Expected output: '{"type": "document", "document": {"id": "your-media-id", "filename": null}}'


from cel.assistants.common import FunctionDefinition, Param



def map_function_parameters_to_tool_message(parameters: list[Param]):
    assert isinstance(parameters, list),\
        "parameters must be a list"

    properties = {}
    for p in parameters:
        properties[p.name] = {
            "type": p.type,
            "description": p.description
        }
        if p.enum:
            properties[p.name]["enum"] = p.enum
            
    return properties


def map_function_to_tool_message(function: FunctionDefinition):
    assert isinstance(function, FunctionDefinition),\
        "function must be an instance of FunctionDefinition"
        
    return {
        "type": "function",
        "function": {
            "name": function.name,
            "description": function.description,
            "parameters": {
                "type": "object",
                "properties": map_function_parameters_to_tool_message(function.parameters),
                "required": [p.name for p in function.parameters if p.required]
            }
        }
    }

def map_functions_to_tool_messages(functions: list[FunctionDefinition]):
    assert isinstance(functions, list),\
        "functions must be a list"
        
    return [map_function_to_tool_message(f) for f in functions]



from typing import Optional
from loguru import logger as log
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import Message
from cel.gateway.model.attachment import LocationAttachment
from cel.decoders.gcp_geocoding import googlev3_decode_location


class GeodecodingMiddleware: 
    """ Middleware to decode meesages with 'location' attachments and add the address to the message text.
    It uses Google Geocoding API to decode the location. 
    """
    
    def __init__(self, location_prefix_msg: str = "My location: "):
        log.debug("GeodecodingMiddleware initialized")
        self.location_prefix_msg = location_prefix_msg
        
    async def __call__(self, 
                       message: Message, 
                       connector: BaseConnector, 
                       assistant: BaseAssistant) -> bool:
        assert isinstance(message, Message), "Message must be a Message object"
        
        
        try:    
            attachments = message.attachments or []
            # has location attachment?
            location_attachment: Optional[LocationAttachment] = next((attachment for attachment in attachments if attachment.type == "location"), None)
            if location_attachment:
                # get the location
                latitude = location_attachment.latitude
                longitude = location_attachment.longitude
                location_coordinates = (latitude, longitude)
                log.debug(f"Location coordinates: {location_coordinates}")
                # get the address
                address = googlev3_decode_location(location_coordinates)
                
                if address:
                    location_attachment.metadata["address"] = address
                    desc = f"({location_attachment.description})" if location_attachment.description else ""
                    
                    # ------------------------------------------------------------------
                    message.text = f"{self.location_prefix_msg} {desc} {address}"
                    # ------------------------------------------------------------------
                    
                    log.debug(f"{self.location_prefix_msg}{location_attachment}")
            return True
        except ValueError as e:
            log.error(f"Error processing Geodecoding: {e}")
            message.text = f"{self.location_prefix_msg}: Error decoding location."
            return True


import json
from loguru import logger as log
from cel.assistants.base_assistant import BaseAssistant
from cel.assistants.macaw.macaw_inference_context import MacawNlpInferenceContext
from cel.assistants.macaw.macaw_nlp import MacawFunctionCall, blend_message, process_insights, process_new_message
from cel.assistants.macaw.macaw_settings import MacawSettings
from cel.gateway.model.conversation_lead import ConversationLead
from cel.prompt.prompt_template import PromptTemplate
from cel.stores.history.base_history_provider import BaseHistoryProvider
from cel.stores.state.base_state_provider import BaseChatStateProvider
from langchain.load.load import loads, load
from langchain.load.dump import dumps


class MacawAssistant(BaseAssistant):
    """Macaw Assistant is a conversational assistant that uses Langchain as core NLP.
    This assistant perform different NLP tasks like blending, insights, and more.
    To perform these tasks, Macaw Assistant uses differents models and tools. You can
    configure the assistant with different settings and tools for each task.
    
    If you use this assistant make sure to setup your LLM API key in the environment variable:
        - OPENAI_API_KEY: The OpenAI API key for the assistant.
    
    Tasks:
        - Blending: Blend a message with the current context.
        - Insights: Get insights from the current context.
        - Process new message: Process a new message with the current context and return the response.

    Args:
        state (dict, optional): Initial state for the assistant. Defaults to None.
        history_store (BaseHistoryProvider, optional): History store provider. Defaults to None.
        state_store (BaseChatStateProvider, optional): State store provider. Defaults to None.
        prompt (PromptTemplate, optional): Prompt template for the assistant. Defaults to None.
        insight_targets (dict, optional): Insight targets for the assistant. Defaults to {}.
        settings (MacawSettings, optional): Settings for the assistant. Defaults to None.

    """
    
    def __init__(self, 
                 state: dict = None,
                 history_store: BaseHistoryProvider = None, 
                 state_store: BaseChatStateProvider = None,
                 prompt: PromptTemplate = None,
                 insight_targets: dict = {},
                 settings: MacawSettings = None,
                 llm=None,
                ):
        
        super().__init__(prompt=prompt, 
                         history_store=history_store, 
                         state_store=state_store)
        
        self.state = state or {}
        self.insight_targets = insight_targets
        if settings is None:
            log.warning("No settings provided for Macaw Assistant, using default settings")
        self.settings = settings or MacawSettings()
        self.llm = llm
        log.debug(f"Macaw Assistant initialized with settings: {self.settings}")
        

    async def new_message(self, lead: ConversationLead, message: str, local_state: dict = {}):
        # create context
        ctx = MacawNlpInferenceContext(
            lead=lead,
            prompt=self.prompt,
            init_state=self.state,
            local_state=local_state,
            functions=self.get_functions(),
            history_store=self._history_store,
            state_store=self._state_store,
            settings=self.settings,
            rag_retriever=self.rag_retriever,
            llm=self.llm
        )
        
        async def on_function_call(ctx: MacawNlpInferenceContext, call: MacawFunctionCall):
            return await self.call_function(
                    func_name=call.name,
                    params=call.args,
                    lead=ctx.lead
                )
        try:
            # stream this message content from string
            async for chunk in process_new_message(ctx, message, on_function_call):
                yield chunk
    
        except Exception as e:
            log.error(f"Macaw Assistant: error processing new message: {e}")
        

    async def blend(self, lead: ConversationLead, text: str, history_length: int = None):
        # create context
        ctx = MacawNlpInferenceContext(
            lead=lead,
            prompt=self.prompt,
            init_state=self.state,
            local_state={},
            history_store=self._history_store,
            state_store=self._state_store,
            settings=self.settings
        )
        
        
        return await blend_message(ctx, message=text)
    
    async def do_insights(self, lead: ConversationLead, targets: dict = {}, history_length: int = 10):
        assert isinstance(targets, dict), "targets must be a dictionary"
        
        if self.settings.insights_enabled is False:
            log.warning("Insights are disabled, returning None")
            return None
        
        log.debug(f"Getting insights for lead: {lead} with targets: {targets}")
        
        mix_targets = {**self.insight_targets, **targets}
        
        # create context
        ctx = MacawNlpInferenceContext(
            lead=lead,
            prompt=self.prompt,
            init_state=self.state,
            local_state={},
            history_store=self._history_store,
            state_store=self._state_store,
            settings=self.settings
        )
        
        insights = await process_insights(ctx, targets=mix_targets)
        
        # raise insight event
        await self.call_event("insigths", lead, data=insights)
        return insights
        
    
    async def process_client_command(self, lead: ConversationLead, command: str, args: list[str]):
    
        if command == "reset":
            
            if args and args[0] == "all":
                await self._history_store.clear_history(lead.get_session_id())
                yield "History cleared"
                await self._state_store.set_store(lead.get_session_id(), {})
                yield "State cleared"
                return

            await self._history_store.clear_history(lead.get_session_id())
            yield "History cleared"
            return
        
        if command == "state":
            state = self._state_store.get_store(lead.get_session_id())
            if state is None:
                yield "No state found"
                return
            for k, v in state.items():
                yield f"{k}: {v}"
            return
        
        if command == "history":
            history = await self._history_store.get_history(lead.get_session_id()) or []    

            if history is None or len(history) == 0:
                yield "History is empty"
                return

            for h in history:
                aux = load(h)
                log.debug(f"History: {aux}")
                yield dumps(aux)
            
            return
        
        if command == "set":
            if len(args) < 2:
                yield "Not enough arguments"
                return
            key = args[0]
            value = args[1]
            state = self._state_store.get_store(lead.get_session_id())
            state[key] = value
            self._state_store.set_store(lead.get_session_id(), state)
            yield f"State updated: {key}: {value}"
            return
        
        if command == "prompt":
            yield "Macaw Assistant v0.1"

            if args and args[0] == "all":            
                prompt = self.prompt
                # split into 250 chars aprox chunks
                for i in range(0, len(prompt), 250):
                    yield prompt[i:i+250]
                return
            
            # first 250 chars
            yield self.prompt[:250]
            return
            
            
            
            


import threading
import time
from loguru import logger as log
from fastapi import APIRouter
from loguru import logger as log
import json
from cel.comms.utils import async_run
from cel.connectors.cli.model.cli_lead import CliLead
from cel.connectors.cli.model.cli_message import CliMessage
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.message_gateway import StreamMode
from cel.gateway.model.message_gateway_context import MessageGatewayContext
from cel.gateway.model.outgoing import OutgoingMessage,\
                                            OutgoingMessageType,\
                                            OutgoingLinkMessage,\
                                            OutgoingSelectMessage,\
                                            OutgoingTextMessage



class CliConnector(BaseConnector):
       
    def __init__(self, stream_mode: StreamMode = StreamMode.SENTENCE):
        log.debug("Creating CLI connector")
        self.router = APIRouter(prefix="/cli")
        self.paused = False
        self.stream_mode = stream_mode

    def cli_listener(self):
        time.sleep(2)
        print("CLI Connector started")
        print("Type 'exit' to stop the connector")
        while True:
            command = input("User: ")
            
            if command == "exit":
                print("Exiting CLI Connector")
                exit()
            
            async_run(self.__process_message(command))
            
            
            

    async def __process_message(self, payload: str):
        try:
            log.debug("Received Cli Message")
            log.debug(payload)
            msg = await CliMessage.load_from_message(payload, connector=self)
            
            if self.paused:
                log.warning("Connector is paused, ignoring message")
                return 
            if self.gateway:
                async for m in self.gateway.process_message(msg, mode=self.stream_mode):
                    pass
            else:
                log.critical("Gateway not set in CLI Connector")
                raise Exception("Gateway not set in CLI Connector")
                
        except Exception as e:
            # TODO: TEST!
            log.error(f"Error processing CLI message: {e}")
                   
    
    async def send_text_message(self, lead: CliLead, text: str, metadata: dict = {}, is_partial: bool = True):
        """ Send a text message to the lead. The simplest way to send a message to the lead.
        
        Args:
            - lead[CliLead]: The lead to send the message
            - text[str]: The text to send
            - metadata[dict]: Metadata to send with the message
            - is_partial[bool]: If the message is partial or not
        """        
        # log.debug(f"Sending message to chat_id: {lead.chat_id}, text: {text}, is_partial: {is_partial}")
        # await self.bot.send_message(chat_id=lead.chat_id, text=text)
        print(f"Bot: {text}")     


    async def send_select_message(self, 
                                  lead: CliLead, 
                                  text: str, 
                                  options: list[str], 
                                  metadata: dict = {}, 
                                  is_partial: bool = True):
        print(f"Bot: {text}")


    async def send_link_message(self, 
                                lead: CliLead, 
                                text: str, 
                                links: list, 
                                metadata: dict = {}, 
                                is_partial: bool = True):
        print(f"Bot: {text}")
        
        
    async def send_typing_action(self, lead: CliLead):
        log.warning(f"Sending typing action to CLI is not implemented yet")
        
        
        
        
    async def send_message(self, message: OutgoingMessage):
        assert isinstance(message, OutgoingMessage),\
            "message must be an instance of OutgoingMessage"
        assert isinstance(message.lead, CliLead),\
            "lead must be an instance of CliLead"
        lead = message.lead
        
        if message.type == OutgoingMessageType.TEXT:
            assert isinstance(message, OutgoingTextMessage),\
            "message must be an instance of OutgoingMessage"
            await self.send_text_message(lead, 
                                         message.content, 
                                         metadata=message.metadata, 
                                         is_partial=message.is_partial)
            
        if message.type == OutgoingMessageType.SELECT:
            assert isinstance(message, OutgoingSelectMessage),\
            "message must be an instance of OutgoingSelectMessage"
            
            await self.send_select_message(lead, 
                                           message.content, 
                                           options=message.options, 
                                           metadata=message.metadata, 
                                           is_partial=message.is_partial)

        if message.type == OutgoingMessageType.LINK:
            assert isinstance(message, OutgoingLinkMessage),\
            "message must be an instance of OutgoingLinkMessage"
            
            await self.send_link_message(lead, 
                                           message.content, 
                                           url=message.links, 
                                           metadata=message.metadata, 
                                           is_partial=message.is_partial)
        
        
        
    def name(self) -> str:
        return "telegram"
        
    def get_router(self) -> APIRouter:
        return self.router
    
    def set_gateway(self, gateway):
        from cel.gateway.message_gateway import MessageGateway
        assert isinstance(gateway, MessageGateway), \
            "gateway must be an instance of MessageGateway"
        self.gateway = gateway
    
    def startup(self, context: MessageGatewayContext):
        log.debug("Starting CLI Connector")
        self.command_thread = threading.Thread(target=self.cli_listener, daemon=True)
        self.command_thread.daemon = True
        self.command_thread.start()
    
    def shutdown(self, context: MessageGatewayContext):
        log.debug("Shutting down telegram connector")
        # async_run(self.bot.delete_webhook(), then=lambda r: log.debug(f'Webhook deleted: {r}'))
        
    def pause(self):
        log.debug("Pausing telegram connector")
        self.paused = True
    
    def resume(self):
        log.debug("Resuming telegram connector")
        self.paused = False



from abc import ABC, abstractmethod
from typing import Sequence, Union
import numpy as np


Embedding = Union[Sequence[float], Sequence[int]]
Embeddings = Sequence[Embedding]



class Text2VectorProvider(ABC):

    @abstractmethod    
    def text2vec(self, text: str) -> Embedding:
        pass

    @abstractmethod
    def texts2vec(self, texts: list[str]) -> Embeddings:
        pass


from abc import ABC
import json

import shortuuid
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.conversation_peer import ConversationPeer


class ConversationLead(ABC):

    def __init__(self, 
                 connector_name: str = None, 
                 metadata : dict = None, 
                 conversation_from: ConversationPeer = None,
                 connector: BaseConnector = None):
        self.connector_name = connector_name or "unknown"
        self.metadata = metadata
        self.conversation_from = conversation_from
        self.tmp_id = shortuuid.uuid()
        self.connector = connector

    def get_session_id(self):
        """ This method should be overriden in the child class """
        return self.tmp_id

    def to_dict(self):
        return {
            'connector_name': self.connector_name,
            'metadata': self.metadata,
            'conversation_from': self.conversation_from.to_dict() if self.conversation_from else None
        }

    def to_json(self):
        return json.dumps(self.to_dict())
    
    @classmethod
    def from_dict(cls, lead_dict):
        return ConversationLead(
            connector_name=lead_dict.get("connector_name"),
            metadata=lead_dict.get("metadata"),
            conversation_from=ConversationPeer.from_dict(lead_dict.get("conversation_from"))
        )
        
    def __str__(self):
        return f"ConversationLead: {self.connector_name}"
    




import json
import threading
import time
import cachetools
import redis as redis_module
from .key_value_store import KeyValueStore

Redis = redis_module.client.Redis
 
class MemRedisCacheAside(KeyValueStore):
    """Implements a cache-aside pattern with memory and Redis cache.

    This class uses both an in-memory cache (using cachetools) and a Redis cache for storing data.
    The in-memory cache is used for fast access, while the Redis cache provides persistence and can be shared across multiple instances.

    Attributes:
        redis_client: A StrictRedis client connected to the Redis server.
        cache: An in-memory LRU cache.
        key_prefix: A prefix added to all keys stored in the Redis cache.
        ttl: The time-to-live (in seconds) for keys in the Redis cache.
        wait_for_redis: A flag indicating whether to wait for Redis operations to complete.
    """

    def __init__(self, redis: str | Redis , key_prefix, memory_maxsize=1000, ttl=60, wait_for_redis=False):
        """Initializes the cache with the given parameters."""
        self.redis_client = redis if isinstance(redis, Redis) else redis_module.StrictRedis.from_url(redis)
        self.cache = cachetools.LRUCache(maxsize=memory_maxsize)
        self.key_prefix = key_prefix
        self.ttl = ttl
        self.wait_for_redis = wait_for_redis

    def set_wait_for_redis(self, wait_for_redis):
        """Sets the wait_for_redis flag. When True, the cache will wait for Redis write operations to complete."""
        self.wait_for_redis = wait_for_redis

    def get(self, key, callback=None):
        """Retrieves the value associated with the given key from the cache.

        If the key is not in the in-memory cache, it tries to retrieve it from the Redis cache.
        If the key is not in the Redis cache either, it calls the provided callback function to compute the value.

        callback: A function that computes the value for the given key if it is not in the cache (optional).
        """        
        data = self.cache.get(key)
        if data is not None:
            return data

        data = self.redis_client.get(self.key_prefix + key)
        if data is not None:
            data = json.loads(data)
            self.cache[key] = data
            return data
        
        if callback:
            data = callback()
            if data is not None:
                self.set(key, data)
                return data

        return None

    def set(self, key, value):
        """Sets the value for the given key in both the in-memory and Redis caches.

        If wait_for_redis is False, the Redis operation is performed in a separate thread. 
        Useful for write-heavy workloads.
        """        
        # Update the in-memory cache
        self.cache[key] = value

        if not self.wait_for_redis:
            # update redis in a separate thread in background
            threading.Thread(target=self.redis_client.setex, args=(self.key_prefix + key, self.ttl, json.dumps(value))).start()
        else:
            #avoid  redis.exceptions.DataError: Invalid input of type: 'dict'. Convert to a bytes, string, int or float first.
            self.redis_client.setex(self.key_prefix + key, self.ttl, json.dumps(value))

    def get_all(self):
        """Returns all items in the in-memory cache."""
        return self.cache.items()

    def delete(self, key):
        """Deletes the key from the in-memory cache."""
        self.cache.pop(key, None)
    
    def delete_redis(self, key):
        """Deletes the key from the Redis cache."""
        self.redis_client.delete(self.key_prefix + key)

    def clear(self):
        """Clears the in-memory cache."""
        self.cache.clear()

    def clear_redis(self):
        """Clears the Redis cache."""
        # delete all keys with the prefix
        keys = self.redis_client.keys(self.key_prefix + '*')
        if keys:
            self.redis_client.delete(*keys)
        


# if __name__ == '__main__':

    # import os
    # from dotenv import dotenv_values
    # config = {
    #     **dotenv_values("example/.env"),    # load development variables
    #     **os.environ,               # override loaded values with environment variables
    # }


    # # Example usage
    # cache = MemRedisCacheAside(
    #     redis_url=config['REDIS_URL'],
    #     key_prefix='cache:',
    #     memory_maxsize=1000,
    #     ttl=60,
    #     wait_for_redis=True
    # )

    # # Set a chat lead into cache
    # lead = gen_event('signature1')
    # id = lead['lead']['signature']
    # cache.set(id, lead)

    # # Get the chat lead from cache
    # l = cache.get(id)
    # # print type of l
    # print(type(l))
    # # print the chat lead id
    # print("Recovered Id:" + l['lead']['signature'])


    # print("Testing performance for wait_for_redis=True")
    # print("---------------------------------------------------")
    # start = time.time()
    # # set 20 keys
    # for i in range(20):
    #     lead = gen_event(f'signature{i}')
    #     id = lead['lead']['signature']
    #     cache.set(id, lead)

    # end = time.time()
    # print(f"Time spent setting 20 keys: {end - start}")

    # print("\n\n\n")


    # print("Testing performance for wait_for_redis=False")
    # print("---------------------------------------------------")
    # cache.set_wait_for_redis(False)
    # start = time.time()
    # # set 20 keys
    # for i in range(20):
    #     lead = gen_event(f'signature{i}')
    #     id = lead['lead']['signature']
    #     cache.set(id, lead)

    # end = time.time()
    # print(f"Time spent setting 20 keys: {end - start}")


from dataclasses import asdict
from loguru import logger as log
import json
import time
from redis import Redis
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import Message
from cel.middlewares.in_mem_blacklist import BlackListEntry



class RedisBlackListMiddleware:
    """Middleware to block users based on a blacklist. The blacklist is stored in a Redis database."""
    
    def __init__(self, redis: str | Redis = None, key_prefix: str = "blacklistmw"):
        self.client = Redis.from_url(redis or 'redis://localhost:6379/0') if isinstance(redis, str) else redis
        self.black_list_key = key_prefix
        
    async def __call__(self, message: Message, connector: BaseConnector, assistant: BaseAssistant):
        assert isinstance(message, Message), "Message must be a Message object"
        
        id =  message.lead.get_session_id()
        source = message.lead.connector_name
        entry = self.client.hget(self.black_list_key, id)
        if entry:
            entry = json.loads(entry)
            log.critical(f"User {id} from {source} is blacklisted. Reason: {entry['reason']}")
            return False
        else:
            return True

    def add_to_black_list(self, id: str, reason: str = None):
        entry = BlackListEntry(reason=reason, date=int(time.time()))
        self.client.hset(self.black_list_key, id, json.dumps(asdict(entry)))
        
    def remove_from_black_list(self, id: str):
        self.client.hdel(self.black_list_key, id)
        
    def get_entry(self, id: str):
        entry = self.client.hget(self.black_list_key, id)
        if entry:
            return json.loads(entry)
        return None




from abc import ABC, abstractmethod
from dataclasses import dataclass
import datetime

from .attachment import FileAttachment
from .conversation_lead import ConversationLead



class Message(ABC):
    """This class represents a generic message object that can be sent or received by the system.
    this class should be subclassed by the specific message platform such as telegram, slack, whatsapp, etc.
    """
    def __init__(self, 
                 lead: ConversationLead, 
                 text: str = None, 
                 date: int = None, 
                 metadata: dict = None, 
                 attachments: list[FileAttachment] = None,
                 isSTT: bool = False,
                 id: str = None
                ):
        self.lead = lead
        self.text = text
        self.metadata = metadata
        self.date = date or datetime.datetime.now().timestamp()
        self.attachments: list[FileAttachment] = attachments
        self.isSTT = isSTT or False
        self.id = id
        
    @abstractmethod
    def is_voice_message(self):
        """This method should be implemented by the subclass to check if the message is a voice message"""
        raise NotImplementedError
    
    @classmethod
    def load_from_dict(cls, message_dict: dict):
        """This method should be implemented by the subclass to load the message from the message object"""
        raise NotImplementedError


import asyncio
from typing import Callable
from loguru import logger as log
from fastapi import APIRouter, Depends, FastAPI, HTTPException, Request, Response
from loguru import logger as log

from cel.assistants.stream_content_chunk import StreamContentChunk
from cel.comms.sentense_detection import streaming_sentence_detector_async
from cel.gateway.model.base_connector import BaseConnector
from cel.assistants.base_assistant import BaseAssistant
from cel.gateway.model.message import ConversationLead, Message 
from cel.gateway.model.message_gateway_context import MessageGatewayContext
from cel.gateway.model.outgoing import OutgoingMessage, OutgoingTextMessage
from cel.message_enhancers.default_message_enhancer import DefaultMessageEnhancer


DEFAULT_CHARS_SLEEP_TIME_RATIO = 25

class StreamMode:
    """
    StreamMode is a class that defines different modes for streaming messages.

    Attributes:
        DIRECT (str): The response is sent as soon as it is ready.
        WORD (str): TODO: not implemented yet.
        SENTENCE (str): pySBD based streaming sentence detector available.
        FULL (str): The whole response is sent at once.
    """
    # the response is sent as soon as it is ready
    DIRECT = "direct"
    # TODO: not implemented yet
    WORD = "word"
    
    # pySBD based streaming sentence detector available
    SENTENCE = "sentence"
    
    # the whole response is sent at once
    FULL = "full"


class MessageGateway:
    """ A message gateway for handling messages from different sources. This class is a singleton.
    MessageGateway is a FastAPI application that can be run as a standalone server. It can be used to
    handle messages from different sources and send them to an assistant for processing.
    
    Since the gateway is a FastAPI application, it can be extended with additional routes and middlewares.
    We strongly recommend using async functions for the middlewares and the routes in connectors.
    Avoid using blocking code in all components of the gateway and the connectors.
    

    Args:
        - webhook_url (str): The webhook url for the gateway. This is the url that the 
        connectors will use to send messages to the gateway. It should be a public url with ssl.
        If you are running the gateway locally, you can use ngrok to create a public url for your local server.
        This is required for connectors that need to receive messages from external sources in a webhook.
        No need to set this if you are using the gateway in a local environment.
        
        - host (str, optional): The host for the gateway. Defaults to '127.0.0.1'.
        
        - port (int, optional): The port for the gateway. Defaults to 8000.
        
        - assistant (BaseAssistant, optional): The assistant to use for processing messages. 
        Defaults to None.
        
        - delivery_rate_control (bool, optional): If True, the gateway will control the delivery rate
        of messages based on the length of the message. Defaults to True. Only available in StreamMode.SENTENCE
        this is useful to prevent the assistant from sending messages too fast to the user. A delay is added
        between messages based on the length of the message.
        
        - delivery_rate_control_ratio (int, optional): The ratio of characters to sleep time. Defaults to 25.
        
        - middlewares (list[Callable[[Message, BaseConnector], bool]], optional): A list of middlewares to
        process messages before they are sent to the assistant. Defaults to None.
        
        - message_enhancer (Callable[[ConversationLead, str, bool], OutgoingMessage], optional): A function
        that enhances the outgoing message before it is sent to the connector. Defaults to None.
        
    """
    
    #singleton
    _instance = None
    
    @staticmethod
    def instance():
        if MessageGateway._instance is None:
            raise Exception("MessageGateway not initialized")
        return MessageGateway._instance
    
    def __init__(self, 
                 assistant: BaseAssistant = None,
                 host: str = '127.0.0.1', 
                 port: int = 8000,
                 webhook_url: str = None,
                 delivery_rate_control: bool = False,
                 delivery_rate_control_ratio: int = DEFAULT_CHARS_SLEEP_TIME_RATIO,
                 middlewares: list[Callable[[Message, BaseConnector], bool]] = None,
                 message_enhancer: Callable[[ConversationLead, str, bool], OutgoingMessage] = None,
                 gateway_api_key: str = None,
                 gateway_api_key_header: str = "x-api-key",
                 *args,
                 **kwargs
                ):
        self.webhook_url = webhook_url or f"http://{host}:{port}"
        self.host = host
        self.port = port
        self.connectors : list[BaseConnector] = []
        self.secured_paths = ["/gateway", "/middlewares"]
        
        
        self.app = FastAPI(
            title="Message Gateway",
            description="A message gateway for handling messages from different sources",
            on_startup=[self.__startup],
            on_shutdown=[self.__shutdown],
            *args, **kwargs
        )
        
        if gateway_api_key is not None:
            def path_is_secure(path: str):
                for p in self.secured_paths:
                    if path.startswith(p):
                        return True
                return False
            
            
            @self.app.middleware("http")
            async def secure_middleware(request: Request, call_next):
                
                if path_is_secure(request.url.path):             
                    if gateway_api_key_header not in request.headers:
                        response = await call_next(request)
                        response.status_code = 401
                        log.error(f"API key not found in headers: {request.url}")
                        log.error(f"Headers: {request.headers}")
                        log.warning(f"Add your API Key into header: {gateway_api_key_header}")
                        return response
                    
                    key = request.headers[gateway_api_key_header]
                    if key != gateway_api_key:
                        response = await call_next(request)
                        response.status_code = 401
                        log.error(f"Invalid API key: {request.url} invalid key: {key}")
                        return response
                
                response = await call_next(request)
                return response
            
        health_router = APIRouter(prefix="/health", tags=["health"])
        
        @health_router.get("/")
        async def health():
            return {"status": "ok"}
        
        
        
        self.app.include_router(health_router)
        self.app.include_router(self.base_routes())
        self.assistant: BaseAssistant = assistant
        self.delivery_rate_control = delivery_rate_control
        self.delivery_rate_control_ratio = delivery_rate_control_ratio
        self.middlewares = middlewares or []
        self.message_enhancer = message_enhancer or DefaultMessageEnhancer()
        
    def register_middleware(self, 
                            middleware: Callable[[Message, BaseConnector, BaseAssistant], bool]):
        self.middlewares.append(middleware)
        log.debug(f"Middleware {middleware.__class__.__name__} registered")
        # if middleware has a setup method, call it, passing self.app as the argument
        # only if the setup method has one argument
        if hasattr(middleware, "setup"):
            try:
                # get middleware class name
                name = middleware.__class__.__name__
                # create a router with prefix: middlewares
                router = APIRouter(prefix=f"/middlewares/{name}", tags=["middlewares"])
                middleware.setup(router)
                # register the router with the app
                self.app.include_router(router)
            except Exception as e:
                log.error(f"Error setting up middleware: {e}")
        
        
        
    def base_routes(self):
       
        router = APIRouter(prefix="/gateway", 
                           tags=["gateway"])
                
        @router.get("/pause")
        async def get_pause():
            for connector in self.connectors:
                connector.pause()
            return {"message": "paused"}
        
        @router.get("/resume")
        async def get_resume():
            for connector in self.connectors:
                connector.resume()
            return {"message": "resumed"}
        
        return router
        
    def __startup(self):
        log.debug("Starting message gateway")
        for connector in self.connectors:
            connector.startup(self.get_context())
    
    def __shutdown(self):
        log.debug("Shutting down message gateway")
        for connector in self.connectors:
            connector.shutdown(self.get_context())

    def get_context(self):
        return MessageGatewayContext(router=APIRouter(), webhook_url=self.webhook_url)
    
    
    async def __process_middlewares(self, message: Message):
        try:
            for middleware in self.middlewares:
                res = await middleware(message, message.lead.connector, self.assistant)
                # Break the chain if any middleware returns False
                if not res:
                    log.error(f"Middleware {type(middleware)} rejected message: {message.text}")
                    return False
            return True
        except Exception as e:
            log.error(f"Error processing middlewares: {e}")
            return False
        
    async def __process_events(self, message: Message):
        try:
            connector = message.lead.connector
            res = await self.assistant.call_event("message", message.lead, message)
            
            if res is None:
                log.debug(f"Event: 'message' response is None for message: {message.text}")
                return True
            
            if res.blend:
                log.debug(f"Blending response: {res.text}")
                text_in_ctx = await self.assistant.blend(message.lead, res.text)
                log.debug(f"Blended response: {text_in_ctx}")
                await connector.send_text_message(message.lead, text_in_ctx)

            if res.disable_ai_response:
                # TODO: implement image, audio, video responses
                if res.text:
                    await connector.send_text_message(message.lead, res.text)
              
                return False
            
        except Exception as e:
            log.error(f"Error processing events: {e}")
            return True
        
    async def __process_client_command(self, message: Message):
        # client commands are commands that are sent by the user to the assistant
        # message.text should be in the format: /command arg1 arg2 arg3
        text = message.text or ''
        if not text.startswith("/"):
            return True
        
        log.warning(f"Got a client command: {message.text}")
        parts = message.text.split(" ")
        command = parts[0][1:]
        args = parts[1:]
        
        # will return false if the command is handled by the assistant implementation
        cmd_executed = not await self.assistant.eval_client_command(
                                                            command=command,
                                                            args=args, 
                                                            message=message)
        
        # if the command was executed by the assistant, return False
        if cmd_executed:
            return False
        
        
        stream = self.assistant.process_client_command(message.lead, command, args)
        async for sentence in stream:
            await message.lead.connector.send_text_message(message.lead, sentence)
            
        return False


        
    async def process_message(self, message: Message, mode: StreamMode = StreamMode.SENTENCE, capture_repsonse: bool = False):
        """Process a message using the assistant. 
        The message is sent to the assistant for processing.
        
        Args:
            - message (Message): The message to process.
            - mode (StreamMode, optional): The mode for streaming the response. Defaults to StreamMode.SENTENCE.
            - capture_repsonse (bool, optional): If True, the response is captured and returned as a generator.
            if False, the response from genAI is sent to the client immediately by the gateway. Defaults to False.
            Look at capture_repsonse as a fire and forget mode when set to False.
            Set to True if you want to capture the response and process it in the connector code. For example,
            if you are building a voice assistant, you can capture the response and use it to synthesize speech.
            
            Fire and Forget use example:
            ```python
            async for chunk in gateway.process_message(message, mode=StreamMode.SENTENCE):
                pass
            ```
            That's it. The gateway will send the response to the user.
            
            Capture response use example:
            ```python
            async for chunk in gateway.process_message(message, mode=StreamMode.SENTENCE, capture_repsonse=True):
                # HERE: process the chunk
                # THEN send the chunk to the user
            ```
        """
        

        assert message is not None, "Message is None"
        assert isinstance(message, Message), "Message is not of type Message"
        # assert mode in [StreamMode.SENTENCE, 
        #                 StreamMode.FULL, 
        #                 StreamMode.DIRECT], "Invalid StreamMode"
        
        
        log.warning(f"Handling message: {message}")
        connector = message.lead.connector
        lead = message.lead
        
        if not await self.__process_middlewares(message):
            log.warning(f"Message {message.lead.get_session_id()} rejected by middlewares")
            return
        
        if not await self.__process_client_command(message):
            return
        
        if not await self.__process_events(message):
            log.warning("Event response has disabled AI response")
            return
        
        await connector.send_typing_action(lead)
        if self.assistant:
            stream = self.assistant.new_message(message.lead, message.text, {})
            content = ''
            
            if mode == StreamMode.SENTENCE:
                async for sentence in streaming_sentence_detector_async(stream): 
                    assert isinstance(sentence, StreamContentChunk), "stream must be a StreamChunk"
                    content += sentence.content
                    
                    if capture_repsonse:
                        yield sentence
                        # pass
                    else:
                        # Send the sentence to the connector
                        await self.dispatch_outgoing_genai_message(message.lead, 
                                                            text=sentence.content, 
                                                            is_partial=sentence.is_partial)
                        await connector.send_typing_action(message.lead)
                        # Time delation based on the length of the sentence
                        if self.delivery_rate_control:
                            length = len(sentence.content)
                            wait_time = length / self.delivery_rate_control_ratio
                            await asyncio.sleep(wait_time)


            if mode == StreamMode.DIRECT:
                async for chunk in stream:
                    assert isinstance(chunk, StreamContentChunk), "stream must be a StreamChunk"
                    content += chunk.content
                    
                    if capture_repsonse:
                        yield chunk
                        # pass
                    else:
                        await self.dispatch_outgoing_genai_message(message.lead, text=chunk.content, is_partial=chunk.is_partial)
                        await connector.send_typing_action(message.lead)

            
            if mode == StreamMode.FULL:
                await connector.send_typing_action(message.lead)
                async for chunk in stream:
                    assert isinstance(chunk, StreamContentChunk), "stream must be a StreamChunk"
                    content += chunk.content
                    
                    if capture_repsonse:
                        yield chunk.content
                        # pass
                    else:
                        await self.dispatch_outgoing_genai_message(message.lead, text=content, is_partial=False)
                
                
                
            log.debug(f"Assistant response: {content}")
        else: 
            log.critical("No assistant available")
            if capture_repsonse:
                yield "No assistant available"
                # pass
            else:
                await connector.send_text_message(message.lead, text="No assistant available")



    @staticmethod
    async def send_text_message(lead: ConversationLead, 
                                text: str, 
                                source_message: Message = None, 
                                is_partial=True):
        
        _self = MessageGateway.instance()
        assert isinstance(lead, ConversationLead), "lead must be an instance of ConversationLead"
        assert isinstance(text, str), "text must be a string"
        assert source_message is None or isinstance(source_message, Message), "source_message must be an instance of Message"
        connector = lead.connector
        
        # await connector.send_text_message(lead, text)
        # create outgoing message
        message = OutgoingTextMessage(content=text, lead=lead, is_partial=is_partial)
        await connector.send_message(message)


    async def dispatch_outgoing_genai_message(self, 
                                              lead: ConversationLead, 
                                              text: str, 
                                              is_partial=False):
        
        assert isinstance(lead, ConversationLead),\
            "lead must be an instance of ConversationLead"
        assert isinstance(text, str),\
            "text must be a string"

        assert isinstance(lead.connector, BaseConnector),\
            "connector must be an instance of BaseConnector"
        connector = lead.connector
        
        # TODO: Call outgoing message enhancer
        msg = await self.message_enhancer(lead, text, is_partial)

        
        # message = OutgoingTextMessage(content=text, lead=lead, is_partial=is_partial)
        await connector.send_message(msg)


    def register_connector(self, connector: BaseConnector):
        assert isinstance(connector, BaseConnector), "Connector must be an instance of BaseConnector"
        conn = connector.get_router()
        assert conn is None or\
            isinstance(conn, APIRouter) or\
            isinstance(conn, FastAPI), "Router must be an instance of APIRouter or FastAPI"
        
        if conn is None:
            log.warning(f"Connector {connector.name()} has no router")
        else:
            if isinstance(conn, FastAPI):
                self.app.mount(f"/{connector.name()}", conn)
            else:
                self.app.include_router(conn)

        self.connectors.append(connector)
        connector.set_gateway(self)

    def run(self, enable_ngrok: bool = False):
        # check if uvicorn is installed
        try:
            import uvicorn
        except ImportError:
            log.error("Uvicorn is not installed. Please install uvicorn to run the gateway")
            return
    
        if enable_ngrok:
            
            try:
                import ngrok
            except ImportError:
                log.error("Ngrok is not installed. \
                          Please install ngrok to run the gateway with ngrok.\
                              Execute the following command to install ngrok:\
                                  pip install ngrok")
                return
            

            log.info("Starting ngrok session")
            async def setup():
                session = await ngrok.SessionBuilder().authtoken_from_env().connect()
                listener = await session.http_endpoint().listen()
                log.info(f"Ngrok public endpoint: {listener.url()}")
                self.webhook_url = listener.url()
                
                # Bind this session to the gateway host and port
                forwardto = f"{self.host}:{self.port}"
                listener.forward(forwardto)
                log.info(f"Forwarding established from {listener.url()} to {forwardto}")

            asyncio.run(setup())
            # uvicorn.run(app=self.app)
            uvicorn.run(self.app, host=self.host, port=self.port)
            return 
                
        
        else:
            uvicorn.run(self.app, host=self.host, port=self.port)
            return
    
        
 


import os
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.conversation_peer import ConversationPeer


class CliLead(ConversationLead):

    def __init__(self, **kwargs):
        super().__init__(connector_name="cli", **kwargs)
        # get current process id from os
        self.process_id = os.getpid()


    def get_session_id(self):
        return f"{self.connector_name}:{self.process_id}"  

    def to_dict(self):
        data = super().to_dict()
        data['process_id'] = self.process_id
        return data

    @classmethod
    def from_dict(cls, lead_dict):
        return CliLead()
        

    def __str__(self):
        return f"CliLead: {self.process_id}"
    
    
    @classmethod
    def from_message(cls, **kwargs):
        conversation_peer = ConversationPeer(
            name='terminal',
            id=os.getpid(),
            phone=None,
            avatarUrl=None,
            email=None
        )
        return CliLead(conversation_from=conversation_peer, **kwargs)


from fastapi import APIRouter
from typing import Any
from .message_gateway_context import MessageGatewayContext


class BaseConnector:

    def name(self) -> str:
        raise NotImplementedError

    def get_router(self) -> APIRouter:
        raise NotImplementedError

    def startup(self, context: MessageGatewayContext):
        raise NotImplementedError

    def shutdown(self, context: MessageGatewayContext):
        raise NotImplementedError

    def pause(self):
        raise NotImplementedError

    def resume(self):
        raise NotImplementedError

    def set_gateway(self, gateway):
        raise NotImplementedError

    async def send_text_message(self, lead, text: str, metadata: dict = {}, is_partial: bool = True):
        raise NotImplementedError
    
    # Send an image message from memory
    async def send_image_message(self, lead, image: Any, filename:str, caption:str = None, metadata: dict = {}, is_partial: bool = True):
        raise NotImplementedError
    
    async def send_image_url_message(self, lead, image_url: str, caption:str = None, metadata: dict = {}, is_partial: bool = True):
        raise NotImplementedError
    
    async def send_typing_action(self, lead):
        raise NotImplementedError
    
    async def send_message(self, message):
        raise NotImplementedError


from dataclasses import dataclass, field
from cel.assistants.function_response import FunctionResponse, RequestMode
from cel.gateway.model.base_connector import BaseConnector
from cel.gateway.model.message import ConversationLead, Message
from cel.stores.history.base_history_provider import BaseHistoryProvider
from cel.stores.state.base_state_provider import BaseChatStateProvider


@dataclass
class FunctionContext:
    lead: ConversationLead
    message: Message | None = None
    connector: BaseConnector | None = None
    state: BaseChatStateProvider | None = None
    history: BaseHistoryProvider | None = None


    @staticmethod
    def response_text(text: str, request_mode: RequestMode = RequestMode.SINGLE):
        return FunctionResponse(text=text, request_mode=request_mode)

    
    # @staticmethod
    # def response_image(image_url: str, request_mode: RequestMode = RequestMode.SINGLE):
    #     return FunctionResponse(image_url=image_url, request_mode=request_mode)


from pathlib import Path

from cel.rag.slicers.markdown.utils import parse_markdown, Block
from ..base_slicer import Slicer, Slice
from loguru import logger as log


class MarkdownSlicer(Slicer):
    """Slice a markdown file. Each slice contains a breadcrumb, header, and text. 
    This slicer seeks for paragraphs, code blocks, blockquotes, tables, and lists.
    Then it creates a slice for each block found. 
    Recommended to use when you need pinpointed slices highly related to the document structure.
    
    For example: 
    
    ```markdown
    # Heading 1
    ## Heading 2
    This is a paragraph.
    ```
    
    Slice content:
    ```
    Heading 1 > Heading 2
    This is a paragraph.
    ```
    
    Breadcumbs are built from the headings found before the current block. 
    Usefull to keep track of the document structure in the slices.
    
    Args:
        file_path: Path to the file to load.

        encoding: File encoding to use. If `None`, the file will be loaded
        with the default system encoding.

        autodetect_encoding: Whether to try to autodetect the file encoding
            if the specified encoding fails.
            
        prefix: Prefix to add to the slice text.
        
        name: Name of the document.
        
        split_table_rows: Whether to split table rows into separate slices. Each slice contains a single breadcrumb+header+row.
        This is useful when you want to treat each row as a separate slice in the RAG model. Default is `False`.
        Recommended to set to `True` when the table is large and you want to avoid large slices.
    """

    def __init__(
        self,
        name: str,
        file_path: str | Path = None,
        content: str | None = None,
        encoding: str = None,
        prefix: str = None,
        split_table_rows: bool = False
    ):
        """Initialize with file path."""
        self.name = name
        self.file_path = file_path
        self.content = content
        self.prefix = prefix
        self.encoding = encoding
        self.split_table_rows = split_table_rows
        
        
    def __load_from_disk(self) -> str:
        
        if self.file_path is None:
            raise ValueError("No file path provided.")
        
        text = ""
        try:
            with open(self.file_path, encoding=self.encoding) as f:
                # read all content
                text = f.read()
                
        except UnicodeDecodeError:
            raise ValueError("Failed to load file with specified encoding.")
        except FileNotFoundError:
            raise ValueError(f"File not found: {self.file_path}")
        except Exception as e:
            raise ValueError(f"Failed to load file: {e}")    
        
            
        return text

    def __load(self) -> str:
        if self.content:
            return self.content
        return self.__load_from_disk()
        
        
    def slice(self) -> list[Slice]:
        slices=[]
        text = self.__load()        
        blocks = parse_markdown(text, split_table_rows=self.split_table_rows)
        
        for i, block in enumerate(blocks):
            log.debug(f"Block: {block}")
            log.debug('------------------------------------------')
        
            # concatenate breadcrumbs in a string
            bc_str = " > ".join(block.breadcrumbs or [])
        
            slice = Slice(
                id= f"{self.name}-{i}",
                text= f"{bc_str}\n{block.text}",
                metadata={
                    'type': block.type, 
                    'name': self.name,
                },
                source=self.name
            )
            slices.append(slice)
            
        return slices


import threading
from diskcache import Cache
import functools
from cel.config import CACHE_DEFAULT_SETTINGS, CACHE_DIRECTORY



def singleton(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if wrapper.instance is None:
            with wrapper.lock:
                if wrapper.instance is None:
                    wrapper.instance = func(*args, **kwargs)
        return wrapper.instance

    wrapper.lock = threading.Lock()
    wrapper.instance = None
    return wrapper

@singleton
def get_cache(directory: str = CACHE_DIRECTORY, **settings):
    return Cache(directory, **{**CACHE_DEFAULT_SETTINGS, **settings})



from abc import ABC
from dataclasses import dataclass


@dataclass
class Param(ABC):
    """ Tool/Function parameter definition, mind that the enum and required 
    fields may not be present in models other than the OpenAI. """
    name: str
    type: str
    description: str
    required: bool = True
    enum: list[str] = None

@dataclass
class FunctionDefinition(ABC):
    name: str
    description: str
    parameters: list[Param]

@dataclass
class EventResponse(ABC):
    text: str = None
    image: str = None
    audio: str = None
    video: str = None
    disable_ai_response: bool = False
    blend: bool = False
    is_private: bool = False



from cel.gateway.model.base_connector import BaseConnector
from cel.connectors.whatsapp.model.whatsapp_attachment import WhatsappAttachment
from cel.connectors.whatsapp.model.whatsapp_lead import WhatsappLead
from cel.gateway.model.conversation_lead import ConversationLead
from cel.gateway.model.message import Message


class WhatsappMessage(Message):
    
    def __init__(self, 
                 lead: ConversationLead, 
                 text: str = None, 
                 metadata: dict = None, 
                 date: int = None,
                 attachments: list[WhatsappAttachment] = None,
                 id: str = None
                ):
        super().__init__(lead, 
                         text=text, 
                         date=date, 
                         metadata=metadata, 
                         attachments=attachments, 
                         id=id)
    
    
    def is_voice_message(self):
        # TODO: Implement voice message check
        return False

    
    
# {
#     "object": "whatsapp_business_account",
#     "entry": [
#         {
#             "id": "103048736088448",
#             "changes": [
#                 {
#                     "value": {
#                         "messaging_product": "whatsapp",
#                         "metadata": {
#                             "display_phone_number": "15550463673",
#                             "phone_number_id": "105602452496989"
#                         },
#                         "contacts": [
#                             {
#                                 "profile": {
#                                     "name": "Alejandro"
#                                 },
#                                 "wa_id": "5491166937848"
#                             }
#                         ],
#                         "messages": [
#                             {
#                                 "from": "5491166937848",
#                                 "id": "wamid.HBgNNTQ5MTE2NjkzNzg0OBUCABIYFDNBQTdBQzAzMDMxOTVBQkQ3QzU5AA==",
#                                 "timestamp": "1717847190",
#                                 "text": {
#                                     "body": "hola"
#                                 },
#                                 "type": "text"
#                             }
#                         ]
#                     },
#                     "field": "messages"
#                 }
#             ]
#         }
#     ]
# }    
    
    
    
    
    @classmethod
    async def load_from_message(cls, 
                                data: dict, 
                                connector: BaseConnector, 
                                token: str,
                                phone_number_id: str = None):
        
        assert token is not None, "token must be provided for attachment download"
        
        assert isinstance(data, dict), "data must be a dictionary"
        assert isinstance(connector, BaseConnector), "connector must be an instance of BaseConnector"
        assert "entry" in data, "data must have an entry key"
        
        entry = data.get("entry")
        assert isinstance(entry, list), "entry must be a list"
        assert len(entry) > 0, "entry must have at least one element"
        
        entry0 = entry[0]
        assert isinstance(entry0, dict), "first entry must be a dictionary"
        assert "changes" in entry0, "first entry must have a changes key"
        
        changes = entry0.get("changes")
        assert isinstance(changes, list), "changes must be a list"
        assert len(changes) > 0, "changes must have at least one element"
        
        change0 = changes[0]
        assert isinstance(change0, dict), "first change must be a dictionary"
        
        change0 = change0.get("value")
        msg0 = change0.get("messages")[0]

        # get text from message or caption if it is a media message
        text = msg0.get("text", {}).get("body") 
        date = int(msg0.get("timestamp"))
        id = msg0.get("id")
        metadata = {}
        lead = WhatsappLead.from_whatsapp_message(data, connector=connector)
        attach = await WhatsappAttachment.load_from_message(data, token, phone_number_id)
        return WhatsappMessage(lead=lead,
                               id=id,
                               text=text,
                               date=date,
                               metadata=metadata,
                               attachments=[attach] if attach else None
                            )
        

    def __str__(self):
        return f"TelegramMessage: {self.text}"
        
    def __repr__(self):
        return f"TelegramMessage: {self.text}"
    
    


import time


def create_chunk_response(id: str, text: str = None):
  res = {
      "id": id,
      "choices": [
          {
              "delta": {
                  "content": text,
                  "function_call": None,
                  "role": None,
                  "tool_calls": None
              },
              "finish_reason": "stop" if text is None else None,
              "index": 0,
              "logprobs": None
          }
      ],
      "created": int(time.time()),
      "model": "gpt-3.5-turbo-0125",
      "object": "chat.completion.chunk",
      "system_fingerprint": None,
      "usage": None
  }
  return res  